class itermethods:
    '''
    Can be mixed in as a superclass to iterables
    to gain its methods.

    Conditions that depend on length first try
    calling len(), then try iterating to detect
    just enough items needed.  Builtin iterables
    often have a fast member variable for their
    length.
    '''
    def one(iterable):
        '''Converts an iterable of one item into its item,
        enforcing that the length must be one.

        Returns:
            Unwrapped item

        Raises:
            KeyError if the container is not of length 1
        '''
        try:
            if len(iterable) != 1:
                raise KeyError('not of length 1')
            for item in iterable:
                return item
        except TypeError:
            found = None
            for item in iterable:
                if found is not None:
                    raise KeyError('multiple items')
                found = item
            if found is None:
                raise KeyError('not found')
            return found
    def first(iterable):
        '''Retrieves the first item from iterating.

        Returns:
            First item found

        Raises:
            KeyError if the iterable is empty
        '''
        for item in iterable:
            return item
        raise KeyError('not found')
    def is_none(iterable):
        '''Returns True if the iterable is empty.'''
        try:
            return len(iterable) == 0
        except TypeError:
            for item in iterable:
                return True
            return False
    def is_one(iterable):
        '''Returns True if the iterable contains exactly 1 item.'''
        try:
            itermethods.one(iterable)
            return True
        except KeyError:
            return False
    def is_many(iterable):
        '''Returns True if the iterable contains more than 1 item.'''
        try:
            return len(iterable) > 1
        except TypeError:
            found = False
            for item in iterable:
                if found:
                    return True
                found = True
            return False



class readonly:
    '''Wraps a container, implying that
       access is read-only.

       The container is proxied, with
       the exception that callable attributes
       are only visible if their names contain
       the string 'get' or they are annotated
       with '@const'.
       '''
    def __init__(self, container):
        self._container = container
    def __getattr__(self, attr):
        val = getattr(self._container, attr)
        if not callable(val):
            return val
        if hasattr(val, '__const') or 'get' in attr:
            setattr(self, attr, val)
            return val
        raise AttributeError(attr)
    def __getitem__(self, item):
        return self._container[item]
    def __iter__(self):
        return iter(self._container)

def const(method):
    '''A decorator that will tell readonly the method is safe.'''
    method.__const = True
    return method



from collections import defaultdict
from smidge import set
from smidge import itermethods
from smidge.readonly import const

# todo? a nice slicing, ellipsis, or kwparams
#       syntax to access values in multiple ways
#       with one method.
class multidict(defaultdict):
    '''
        -> new dictionary object

        A multidict is a key-value mapping where
        each key can map to multiple values.

        Parameters:
            mapping_or_iterable:
                An optional object to take initial (key,value) tuples from.
            default_factory:
                A container to use for multiple values.
            mapping:
                Optional key=value pairs to add.
                If the type of a value is default_factory, it will be expanded.
    '''
    def __init__(self, mapping_or_iterable = (), default_factory = set, **mapping):
        super().__init__(default_factory)
        try:
            mapping_or_iterable = mapping_or_iterable.items()
        except:
            pass
        for key, value in mapping_or_iterable:
            self.add(key, value)
        for key, value in mapping.items():
            if type(value) is default_factory:
                self.add(key, *value)
            else:
                self.add(key, value)
    @const
    def all(self, key):
        '''Return the container of all items matching the given key.'''
        return self[key]
    @const
    def one(self, key):
        '''Return the single item matching a given key.

        Raises KeyError if there is not a unique item matching the key.
        '''
        return itermethods.one(self.all(key))
    @const
    def first(self, key):
        '''Return one item that matches the given key.

        Raises KeyError if no items match.
        '''
        return itermethods.first(self.all(key))
    @const
    def len(self, key):
        '''Returns the number of items matching the given key.'''
        return len(self.all(key))
    def add(self, key, *values):
        '''Adds values to a key in the multidict.'''
        container = self.all(key)
        for value in values:
            container.add(value)
    def discard(self, key, *values):
        '''Removes values from the multidict.'''
        oldset = self.all(key)
        for value in values:
            oldset.discard(value)
        # delete if set is now empty
        for item in oldset:
            return
        del self[key]
    def move(self, oldkey, newkey, *values):
        '''Changes the key associated with values.

        None may be provided for a key, to create or remove values with generic code.'''
        if oldkey != newkey:
            if oldkey is not None:
                self.discard(oldkey, *values)
            if newkey is not None:
                self.add(newkey, *values)
    @const
    def values(self):
        for container in super().values():
            yield from container
    @const
    def __len__(self):
        sum = 0
        for container in super().values():
            sum += len(container)
        return sum
    def __missing__(self, key):
        value = self.default_factory()
        super().__setitem__(key, value)
        return value
    def __setitem__(self, key, value):
        container = self.all(key)
        if type(value) is type(container):
            container.clear()
            container.update(value)
        else:
            if itermethods.is_many(container):
                raise KeyError('multiple items')
            container.clear()
            container.add(value)



from smidge.itermethods import itermethods
from smidge.readonly import readonly, const # todo? const's object-labeling behavior could be generic

class set(__builtins__['set'], itermethods):
    '''
    set() -> new empty set
    set(*iterable) -> new set of many items
    set(item) -> new set of 1 item
    set(item1, item2, item3) -> new set of 3 items

    Build a collection of unique elements.
    '''
    def __init__(set, *params):
        super().__init__(params)

from smidge.multidict import multidict




import setuptools

setuptools.setup(
    name='smidge',
    version='0.0.4',
    descrption='Small python tools',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/xloem/pysmidge',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    packages=[
        'smidge'
    ],
    python_requires='>=3.7'
)
